@@ -1,8 +1,8 @@
 /*
 
-  grepcidr 1.3 - Filter IP addresses matching IPv4 CIDR specification
-  Copyright (C) 2004, 2005  Jem E. Berkes <jberkes@pc-tools.net>
-  	www.sysdesign.ca
+  grepcidr 1.4 - Filter IP addresses matching IPv4 CIDR specification
+    Copyright (C) 2004 - 2013  Jem E. Berkes <jem@berkes.ca>
+  	www.berkes.ca
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -25,17 +25,19 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include "getopt.h"
+#include <unistd.h>
+#include <sys/types.h>
+#include <regex.h>
 
 #define EXIT_OK		0
 #define EXIT_NOMATCH	1
 #define EXIT_ERROR	2
 
-#define TXT_VERSION	"grepcidr 1.3\nCopyright (C) 2004, 2005  Jem E. Berkes <jberkes@pc-tools.net>\n"
+#define TXT_VERSION	"grepcidr 1.4\nCopyright (C) 2004 - 2013  Jem E. Berkes <jem@berkes.ca>\n"
 #define TXT_USAGE	"Usage:\n" \
-			"\tgrepcidr [-V] [-c] [-v] PATTERN [FILE]\n" \
-			"\tgrepcidr [-V] [-c] [-v] [-e PATTERN | -f FILE] [FILE]\n"
-#define MAXFIELD	512
+			"\tgrepcidr [-V] [-c] [-v] [-x] PATTERN [FILE]\n" \
+			"\tgrepcidr [-V] [-c] [-v] [-x] [-e PATTERN | -f FILE] [FILE]\n"
+#define MAXFIELD	1024
 #define TOKEN_SEPS	"\t,\r\n"	/* so user can specify multiple patterns on command line */
 #define INIT_NETWORKS	8192
 
@@ -57,6 +59,9 @@ struct netspec
 
 
 /* Global variables */
+int anymatch = 0;				/* did anything match? for exit code */
+int invert = 0;				/* flag for inverted mode */
+unsigned int counting = 0;		/* when non-zero, counts matches */
 unsigned int patterns = 0;		/* total patterns in array */
 unsigned int capacity = 0;		/* current capacity of array */
 struct netspec* array = NULL;		/* array of patterns, network specs */
@@ -153,17 +158,40 @@ int netsearch(const void* a, const void* b)
 }
 
 
+/*
+	Search for this IP (passed in key) among loaded network patterns and determine if it's a "match"
+	Returns 0 if no match, no line/count is output
+	Returns nonzero if match, line/count is output
+*/
+int match_ip(struct netspec* key, char* line)
+{
+	int match = 0;
+	if (bsearch(key, array, patterns, sizeof(struct netspec), netsearch))
+		match = 1;
+	if (invert ^ match)
+	{
+		anymatch = 1;
+		if (counting)
+			counting++;
+		else
+			printf("%s", line);
+		return 1;	/* match */
+	}
+	else
+		return 0;	/* no match */
+}
+
+
 int main(int argc, char* argv[])
 {
-	static char shortopts[] = "ce:f:vV";
+	static char shortopts[] = "ce:f:vxV";
 	FILE* inp_stream = stdin;		/* input stream, list of IPs to match */
 	char* pat_filename = NULL;		/* filename containing patterns */
 	char* pat_strings = NULL;		/* pattern strings on command line */
-	unsigned int counting = 0;		/* when non-zero, counts matches */
-	int invert = 0;				/* flag for inverted mode */
 	char line[MAXFIELD];
+	int strict_noregex = 0;			/* flag for original style, single IP match */
 	int foundopt;
-	int anymatch = 0;			/* did anything match? for exit code */
+	static regex_t preg;			/* compiled regular expression for IPs */
 
 	if (argc == 1)
 	{
@@ -187,6 +215,10 @@ int main(int argc, char* argv[])
 				invert = 1;
 				break;
 				
+			case 'x':
+				strict_noregex = 1;
+				break;
+				
 			case 'e':
 				pat_strings = optarg;
 				break;
@@ -272,22 +304,35 @@ int main(int argc, char* argv[])
 		}
 	}
 	
+	/* Compile the regular expression for matching IP addresses */
+	if (regcomp(&preg, "[0-9]+\\.[0-9]+\\.[0-9]+\\.[0-9]+", REG_EXTENDED) != 0)
+	{
+		fputs("regcomp() failed\n", stderr);
+		return EXIT_ERROR;
+	}
+
 	/* Match IPs from input stream to network patterns */
 	while (fgets(line, sizeof(line), inp_stream))
 	{
 		struct netspec key;
-		if ((key.min=ip_to_uint(line)))
+		if (strict_noregex)
 		{
-			int match = 0;
-			if (bsearch(&key, array, patterns, sizeof(struct netspec), netsearch))
-				match = 1;
-			if (invert ^ match)
+			/* Faster old-style search, look for single IP from start of line */
+			if ((key.min=ip_to_uint(line)))
+				match_ip(&key, line);
+		}
+		else
+		{
+			/* Regex search, find IP anywhere on the line */
+			regoff_t offset;
+			regmatch_t pmatch;
+			for (offset = 0; regexec(&preg, &line[offset], 1, &pmatch, 0) == 0; offset += pmatch.rm_eo)
 			{
-				anymatch = 1;
-				if (counting)
-					counting++;
-				else
-					printf("%s", line);
+				if ((key.min=ip_to_uint(&line[offset + pmatch.rm_so])))
+				{
+					if (match_ip(&key, line))
+						break;	/* this line is a match, stop scanning inside line */
+				}
 			}
 		}
 	}
@@ -297,6 +342,7 @@ int main(int argc, char* argv[])
 		fclose(inp_stream);
 	if (array)
 		free(array);
+	regfree(&preg);
 
 	if (counting)
 		printf("%u\n", counting-1);
